/*
 * Commands to compile and run
 *
 * g++ -o hello -lX11 hello.C
 * ./hello
 *
 *
 * On cygwin, you need to give the location of the X11 libraries,
 * something like (depending on your installation):
 * 
 * g++ -o hello hello.C -L/cygdrive/c/cygwin/usr/X11R6/lib/ -lX11
 *
 * Also note that in "-lX11" the 2nd character is a lower case L while
 * the fourth and fifth are ones.
 */
#include <iostream>
#include <list>

/*
 * Header files for X functions
 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>

using namespace std;

/*
 * Defines for window geometry and keyboard events
 */
 
const int Border = 5;
const int BufferSize = 10;


/*
 * Information to draw on the window.
 */
struct XInfo
{  Display  *display;
   Window   window;
   GC       gc;
};


/*
 * An abstract class representing displayable things. 
 */
class Displayable
{ public:
      virtual void paint(XInfo &xinfo) = 0;
};       

/*
 * Display some text where the user clicked the mouse.
 */
class Text : public Displayable
{  public:
      virtual void paint(XInfo &xinfo)
      {  XDrawImageString( xinfo.display, xinfo.window, xinfo.gc, 
               this->x, this->y, this->s.c_str(), this->s.length() );
      }
      
      // constructor
      Text(int x, int y, string s):x(x), y(y), s(s)  {}
      
   private:
      int x;
      int y;
      string s;
};

/*
 * Function to put out a message on error exits.
 */
void error( string str ) {
  cerr << str << endl;
  exit(0);
}

/*
 * Function to repaint a display list
 */
void repaint( list<Displayable *> dList, XInfo &xinfo) 
{
   list<Displayable *>::const_iterator begin = dList.begin();
   list<Displayable *>::const_iterator end = dList.end();

   XClearWindow( xinfo.display, xinfo.window );
   while( begin != end ) 
   {  Displayable *d = *begin;
      d->paint(xinfo);
      begin++;
   }
   XFlush( xinfo.display );
}

/*
 * The loop responding to events from the user. 
 */
void eventloop(XInfo &xinfo)
{  XEvent event;
   KeySym key;
   char text[BufferSize];
   list<Displayable *> dList;           // list of Displayables

   while( true ) 
   {  XNextEvent( xinfo.display, &event );
      switch( event.type ) {
      /*
       * Repaint the window on expose events.
       */
      case Expose:
         if ( event.xexpose.count == 0 ) 
         {  repaint( dList, xinfo);
         }
         break;
      /*
       * Add an item where the mouse is clicked, and repaint.
       */
      case ButtonPress:
         dList.push_front(new Text(event.xbutton.x, event.xbutton.y, "Urrp!"));
         repaint( dList, xinfo );
         break;
      /*
       * Exit when 'q' is typed.
       * Arguments for XLookupString :
       *                 event - the keyboard event
       *                 text - buffer into which text will be written
       *                 BufferSize - size of text buffer
       *                 key - workstation independent key symbol
       *                 0 - pointer to a composeStatus structure
       */
      case KeyPress:
         int i = XLookupString( (XKeyEvent *)&event, text, BufferSize, &key, 0 );
         if ( i == 1 && text[0] == 'q' )
         {  error( "Terminated normally." );
         }
         break;
      }
   }
}

/*
 * Create the window;  initialize X.
 */
void initX(int argc, char *argv[], XInfo &xinfo)
{  XSizeHints hints;
   unsigned long background, foreground;
   int screen;

   /*
    * Display opening uses the DISPLAY  environment variable.
    * It can go wrong if DISPLAY isn't set, or you don't have permission.
    */  
   xinfo.display = XOpenDisplay( "" );
   if ( !xinfo.display )
   { error( "Can't open display." );
   }

   /*
    * Find out some things about the display you're using.
    */
   screen = DefaultScreen( xinfo.display );
   background = WhitePixel( xinfo.display, screen );
   foreground = BlackPixel( xinfo.display, screen );

   /*
    * Set up hints and properties for the window manager, and open a window.
    * Arguments to XCreateSimpleWindow :
    *                 display - the display on which the window is opened
    *                 parent - the parent of the window in the window tree
    *                 x,y - the position of the upper left corner
    *                 width, height - the size of the window
    *                 Border - the width of the window border
    *                 foreground - the colour of the window border
    *                 background - the colour of the window background.
    * Arguments to XSetStandardProperties :
    *                 display - the display on which the window exists
    *                 window - the window whose properties are set
    *                 Hello1 - the title of the window
    *                 Hello2 - the title of the icon
    *                 none - a pixmap for the icon
    *                 argv, argc - a comand to run in the window
    *                 hints - sizes to use for the window.
    */
   hints.x = 100;
   hints.y = 100;
   hints.width = 400;
   hints.height = 300;
   hints.flags = PPosition | PSize;
   xinfo.window = XCreateSimpleWindow( xinfo.display, DefaultRootWindow( xinfo.display ),
                                hints.x, hints.y, hints.width, hints.height,
                                Border, foreground, background );
   XSetStandardProperties( xinfo.display, xinfo.window, "Hello1", "Hello2", None,
                                argv, argc, &hints );

   /*
    * Get a graphics context and set the drawing colours.
    * Arguments to XCreateGC
    *           display - which uses this GC
    *           window - which uses this GC
    *           GCflags - flags which determine which parts of the GC are used
    *           GCdata - a struct that fills in GC data at initialization.
    */
   xinfo.gc = XCreateGC (xinfo.display, xinfo.window, 0, 0 );
   XSetBackground( xinfo.display, xinfo.gc, background );
   XSetForeground( xinfo.display, xinfo.gc, foreground );

   /*
    * Tell the window manager what input you want.
    */
   XSelectInput( xinfo.display, xinfo.window,
                ButtonPressMask | KeyPressMask | ExposureMask );

   /*
    * Put the window on the screen.
    */
   XMapRaised( xinfo.display, xinfo.window );
}



/*
 * Start executing here.
 *   First initialize window.
 *   Next loop responding to events.
 *   Exit forcing window manager to clean up - cheesy, but easy.
 */
int main ( int argc, char *argv[] ) 
{  XInfo xinfo;

   initX(argc, argv, xinfo);
   eventloop(xinfo);
}
